// An incrementally updatable tree for class browsing.
// See also: ValueTree
// 24 Jul 1997 Derek Lieber
//
//
package jd;
import  java.util.*;
import  java.util.zip.*;
import  java.io.*;

class ClassTree implements Constants
   {
   private Vector repositories; // places to look for classes
   private Vector trees;        // tree representation of classes found so far
   private String text;         // printable version of tree (null --> must regenerate)
   
   ClassTree(Vector classRepositories)
      {
      this.repositories = classRepositories;
      this.trees        = new Vector();
      
      for (Enumeration e = repositories.elements(); e.hasMoreElements(); )
         {
         String repositoryName = (String)e.nextElement();

         try
            {
            if (repositoryName.endsWith(".zip") || repositoryName.endsWith(".jar"))
               trees.addElement(new ZipTree(repositoryName, repositoryName));
            else
               trees.addElement(new FileTree(null, repositoryName, (new File(repositoryName)).getCanonicalPath()));
            }

         catch(IOException x)
            {
            Session.popup("classTree: " + x);
            }
         
         text = null;
         renumber();
         }
      }

   synchronized
   void
   expandOrCollapse(int targetLino)
      {
      for (Enumeration e = trees.elements(); e.hasMoreElements(); )
         {
         AnyTree tree = (AnyTree)e.nextElement();
         if (tree.expandOrCollapse(targetLino))
            break;
         }
      text = null;
      renumber();
      }

   synchronized
   private void
   renumber()
      {
      int lino = 0;
      for (Enumeration e = trees.elements(); e.hasMoreElements(); )
         {
         AnyTree tree = (AnyTree)e.nextElement();
         lino = tree.renumber(lino);
         }
      }

   synchronized
   String
   getText()
      {
      if (text == null)
         {
      // System.out.println("class tree: print");
         StringBuffer buf = new StringBuffer();
         for (Enumeration e = trees.elements(); e.hasMoreElements(); )
            {
            AnyTree tree = (AnyTree)e.nextElement();
            tree.print(buf, "");
            }
         text = buf.toString();
         }
      return text;
      }

   synchronized
   String
   linoToName(int targetLino)
      {
   // System.out.println("class tree: linoToName");
      for (Enumeration e = trees.elements(); e.hasMoreElements(); )
         {
         AnyTree tree   = (AnyTree)e.nextElement();
         String  target = tree.linoToName(targetLino);
         if (target != null)
            {
         // System.out.println("class tree: linoToName(" + targetLino + ") --> " + target);
            return target;
            }
         }
   // System.out.println("class tree: linoToName(" + targetLino + ") --> " + null);
      return null;
      }

   synchronized
   int
   nameToLino(String targetName)
      {
   // System.out.println("class tree: nameToLino");

      // search each tree for class.
      //
      for (Enumeration e = trees.elements(); e.hasMoreElements(); )
         {
         AnyTree tree   = (AnyTree)e.nextElement();
         int     target = tree.nameToLino(targetName);
         if (target != -1)
            {
         // System.out.println("class tree: nameToLino(" + targetName + ") --> " + target);
            return target;
            }
         }

      // class wasn't found, perhaps because tree hasn't yet been expanded along path required to reach it.
      // let's expand each tree along path required to reach class (if it exists).
      //
      for (Enumeration e = trees.elements(); e.hasMoreElements(); )
         {
         AnyTree tree = (AnyTree)e.nextElement();
         tree.expandToReach(targetName);
         }
      
      text = null;
      renumber();
      
      // now search again
      //
      for (Enumeration e = trees.elements(); e.hasMoreElements(); )
         {
         AnyTree tree   = (AnyTree)e.nextElement();
         int     target = tree.nameToLino(targetName);
         if (target != -1)
            {
         // System.out.println("class tree: nameToLino(" + targetName + ") --> " + target);
            return target;
            }
         }

   // System.out.println("class tree: nameToLino(" + targetName + ") --> -1");
      return -1;
      }
   }

interface AnyTree
   {
   String  descriptiveName();
   String  linoToName(int targetLino);
   int     nameToLino(String targetName);
   boolean expandOrCollapse(int targetLino);
   boolean expandToReach(String targetName);
   int     renumber(int counter);
   void    print(StringBuffer out, String indent);
   }
   
class FileTree implements Constants, AnyTree
   {
   int      lino;             // line number for this tree
   FileTree parent;           // enclosing directory (null --> this is root)
   String   descriptiveName;  // descriptive identifier for this directory - something like "bar"
   String   canonicalDirName; // canonicalized name of this directory      - something like "d:\foo\bar"
   Vector   classes;          // fully qualified names of classes residing in this directory (element type == String)
   Vector   subtrees;         // subdirectories of this directory                            (element type == FileTree)

   FileTree(FileTree parent, String descriptiveName, String canonicalDirName)
      {
      this.lino             = 0;
      this.parent           = parent;
      this.descriptiveName  = descriptiveName;
      this.canonicalDirName = canonicalDirName;
      this.classes          = null;
      this.subtrees         = null;
      }

   // Get one-word description of this tree.
   //
   synchronized
   public String
   descriptiveName()
      {
      return descriptiveName;
      }
      
   // Get classname appearing at specified line of tree.
   // Taken:    line number (0-indexed)
   // Returned: classname appearing at that line (null --> not found)
   //
   synchronized
   public String
   linoToName(int targetLino)
      {
      int counter = lino;

      if (counter == targetLino)
         return null;

      if (classes != null)
         for (Enumeration e = classes.elements(); e.hasMoreElements(); )
            {
            String className = (String)e.nextElement();
            counter += 1;
            if (counter == targetLino)
               return className;
            }

      if (subtrees != null)
         for (Enumeration e = subtrees.elements(); e.hasMoreElements(); )
            {
            FileTree tree = (FileTree)e.nextElement();
            String target = tree.linoToName(targetLino);
            if (target != null)
               return target;
            }

      return null;
      }

   // Get line number at which specified classname appears in tree.
   // Taken:    classname
   // Returned: line number at which that name appears (-1 --> not found)
   //
   synchronized
   public int
   nameToLino(String targetName)
      {
      int counter = lino;

      if (classes != null)
         for (Enumeration e = classes.elements(); e.hasMoreElements(); )
            {
            String className = (String)e.nextElement();
            counter += 1;
            if (className.equals(targetName))
               return counter;
            }

      if (subtrees != null)
         for (Enumeration e = subtrees.elements(); e.hasMoreElements(); )
            {
            FileTree tree = (FileTree)e.nextElement();
            int    target = tree.nameToLino(targetName);
            if (target != -1)
               return target;
            }

      return -1;
      }

   // Expand or collapse subtree at specified line number.
   // Taken:    line number sought
   // Returned: was there a subtree at that line?
   //
   synchronized
   public boolean
   expandOrCollapse(int targetLino)
      {
      if (lino == targetLino)
         {
         if (subtrees == null)
            { // expand detail
            classes  = new Vector();
            subtrees = new Vector();
            populate();
            }
         else
            { // collapse detail
            classes  = null;
            subtrees = null;
            }
         return true;
         }

      if (subtrees != null)
         for (Enumeration e = subtrees.elements(); e.hasMoreElements(); )
            {
            FileTree tree = (FileTree)e.nextElement();
            if (tree.expandOrCollapse(targetLino))
               return true;
            }

      return false;
      }

   // Expand tree along path required to reach specified class (if it exists).
   // Taken:    fully qualified name of class sought
   // Returned: was tree fully expanded along path required to reach that class?
   //
   synchronized
   public boolean
   expandToReach(String targetName)
      {
      int    dotIndex      = targetName.lastIndexOf('.');
      String targetPrefix  = (dotIndex == -1) ? "" : targetName.substring(0, dotIndex);
      String targetDirName = targetPrefix.replace('.', FILE_PATH_SEPARATOR);

      String canonicalRootName = canonicalDirName;
      for (FileTree t = this; t.parent != null; t = t.parent)
         canonicalRootName = t.parent.canonicalDirName;

      String relativeDirName = canonicalRootName.equals(canonicalDirName)
                             ? ""
                             : canonicalDirName.substring(canonicalRootName.length() + 1, canonicalDirName.length());

      // an example:
      //
      // targetName:         ddd.eee.fff
      // targetDirName:      ddd\eee
      // canonicalRootName:  d:\aaa\bbb\ccc
      // canonicalDirName:   d:\aaa\bbb\ccc\ddd\eee
      // relativeDirName:    ddd\eee

      if (!targetDirName.startsWith(relativeDirName))
         return false;
         
      if (classes == null)
         { // expand this tree
      // System.out.println("expandingToReach " + targetName + " from " + canonicalDirName);
         classes  = new Vector();
         subtrees = new Vector();
         populate();
         }
         
      if (relativeDirName.equals(targetDirName))
         return true; // reached final path component
         
      // expand appropriate subtree
      for (Enumeration e = subtrees.elements(); e.hasMoreElements(); )
         {
         FileTree tree = (FileTree)e.nextElement();
         if (tree.expandToReach(targetName))
            return true;
         }

      return false;
      }
      
   // Populate "classes" with this directory's classes and "subtrees" with this directory's subdirectories.
   //
   synchronized
   void
   populate()
      {
      File   dir     = new File(canonicalDirName);
      String files[] = dir.list();

      if (files == null)
         return;

   // System.out.println("populating classtree from " + canonicalDirName);

      String canonicalRootName = canonicalDirName;
      for (FileTree t = this; t.parent != null; t = t.parent)
         canonicalRootName = t.parent.canonicalDirName;                                      // "d:\xxx"

      for (int i = 0, n = files.length; i < n; ++i)
         {
         String simpleFileName    = files[i];                                                // "zzz.class"
         String canonicalFileName = canonicalDirName + FILE_PATH_SEPARATOR + simpleFileName; // "d:\xxx\yyy\zzz.class"

         int tailIndex = canonicalFileName.lastIndexOf(".class");
         if (tailIndex > 0)
            {
            String tmp = canonicalFileName.substring(canonicalRootName.length() + 1,         // remove "d:\xxx\"
                                                     tailIndex);                             // remove ".class"
            String className = tmp.replace(FILE_PATH_SEPARATOR, '.');                        // change "\" to "."
            classes.addElement(new String(className));                                       // "yyy.zzz"
         // System.out.println("populate: fileName=" + canonicalFileName + " className=" + className);
            continue;
            }

         File fileOrDirectory = new File(canonicalFileName);
         if (fileOrDirectory.isDirectory())
            subtrees.addElement(new FileTree(this, simpleFileName, canonicalFileName));
         }
      }

   // Renumber tree nodes after expanding/collapsing tree.
   //
   synchronized
   public int
   renumber(int counter)
      {
      lino = counter++;

      if (subtrees != null)
         {
         counter += classes.size();
         for (Enumeration e = subtrees.elements(); e.hasMoreElements(); )
            {
            FileTree tree = (FileTree)e.nextElement();
            counter = tree.renumber(counter);
            }
         }

      return counter;
      }

   // Generate textual description of this tree's contents.
   //
   synchronized
   public void
   print(StringBuffer out, String indent)
      {
      out.append(indent);
      out.append('-');
      out.append(descriptiveName);
      out.append('\n');

      indent += " ";

      if (classes != null)
         for (Enumeration e = classes.elements(); e.hasMoreElements(); )
            {
            String className = (String)e.nextElement();
            out.append(indent);
            out.append(className);
            out.append('\n');
            }

      if (subtrees != null)
         for (Enumeration e = subtrees.elements(); e.hasMoreElements(); )
            {
            FileTree tree = (FileTree)e.nextElement();
            tree.print(out, indent);
            }
      }
   }
   
class ZipTree implements AnyTree
   {
   int      lino;             // line number for this tree
   String   zipArchiveName;   // zip archive from which this tree is built
   String   descriptiveName;  // descriptive identifier for this tree
   Vector   classes;          // fully qualified names of classes residing in this tree (element type == String)

   ZipTree(String zipArchiveName, String descriptiveName)
      {
      this.zipArchiveName   = zipArchiveName;
      this.descriptiveName  = descriptiveName;
      this.classes          = null;
      }

   // Get one-word description of this tree.
   //
   synchronized
   public String
   descriptiveName()
      {
      return descriptiveName;
      }
   
   // Get classname appearing at specified line of tree.
   // Taken:    line number (0-indexed)
   // Returned: classname appearing at that line (null --> not found)
   //
   synchronized
   public String
   linoToName(int targetLino)
      {
      int counter = lino;

      if (counter == targetLino)
         return null;

      if (classes != null)
         for (Enumeration e = classes.elements(); e.hasMoreElements(); )
            {
            String className = (String)e.nextElement();
            counter += 1;
            if (counter == targetLino)
               return className;
            }

      return null;
      }
      
   // Get line number at which specified classname appears in tree.
   // Taken:    classname
   // Returned: line number at which that name appears (-1 --> not found)
   //
   synchronized
   public int
   nameToLino(String targetName)
      {
      int counter = lino;

      if (classes != null)
         for (Enumeration e = classes.elements(); e.hasMoreElements(); )
            {
            String className = (String)e.nextElement();
            counter += 1;
            if (className.equals(targetName))
               return counter;
            }

      return -1;
      }
      
   // Expand or collapse tree at specified line number.
   // Taken:    line number sought
   // Returned: was there a subtree at that line?
   //
   synchronized
   public boolean
   expandOrCollapse(int targetLino)
      {
      if (lino == targetLino)
         {
         if (classes == null)
            { // expand detail
            classes  = new Vector();
            try
               {
               populate();
               }
            catch(IOException x)
               {
               Session.popup("classTree: " + x);
               }
            }
         else
            { // collapse detail
            classes  = null;
            }
         return true;
         }

      return false;
      }

   // Expand tree along path required to reach specified class.
   // Taken:    fully qualified name of class sought
   // Returned: could tree be fully expanded along path required to reach that class?
   //
   synchronized
   public boolean
   expandToReach(String targetName)
      {
   // System.out.println("expandingToReach " + targetName + " from " + zipArchiveName);
      if (classes == null)
         expandOrCollapse(lino);
      return true;
      }
      
   // Populate "classes" with this zip archive's classes.
   //
   synchronized
   void
   populate()
      throws IOException
      {
   // System.out.println("populating classtree from " + zipArchiveName);
      
      // extract class names from archive
      //
      ZipFile archive = new ZipFile(zipArchiveName);
      for (Enumeration e = archive.entries(); e.hasMoreElements(); )
         {
         ZipEntry entry = (ZipEntry)e.nextElement();
         String   canonicalFileName = entry.getName();

         int tailIndex = canonicalFileName.lastIndexOf(".class");
         if (tailIndex > 0)
            {
            String tmp = canonicalFileName.substring(0, tailIndex); // remove ".class"
            String className = tmp.replace('/', '.');               // change "/" to "."
            classes.addElement(new String(className));              // "yyy.zzz"
            continue;
            }
         }
      archive.close();

      // alphabetize (K&R shell sort)
      //
      for (int n = classes.size(), gap = n / 2; gap > 0; gap /= 2)
         for (int i = gap; i < n; ++i)
            for (int j = i - gap; j >= 0; j -= gap)
               {
               String a = (String)classes.elementAt(j);
               String b = (String)classes.elementAt(j + gap);

               // a < b?
               if (a.compareTo(b) < 0)
                  break;

               // swap
               Object tmp = classes.elementAt(j);
               classes.setElementAt(classes.elementAt(j + gap), j);
               classes.setElementAt(tmp, j + gap);
               }
      }

   // Renumber tree nodes after expanding/collapsing a subtree.
   //
   synchronized
   public int
   renumber(int counter)
      {
      lino = counter++;

      if (classes != null)
         counter += classes.size();

      return counter;
      }

   // Generate textual description of this tree's contents.
   //
   synchronized
   public void
   print(StringBuffer out, String indent)
      {
      out.append(indent);
      out.append('-');
      out.append(descriptiveName);
      out.append('\n');

      indent += " ";

      if (classes != null)
         {
         for (Enumeration e = classes.elements(); e.hasMoreElements(); )
            {
            String className = (String)e.nextElement();

            out.append(indent);
            out.append(className);
            out.append('\n');
            }
         }
      }
   }
