// Debugger state repository and execution controller.
// 11 Mar 1996 Derek Lieber
//
// We run the show using three threads.
//
//       name                         purpose
//  --------------   ----------------------------------------------
//  "Agent input"    enqueue events reported by remote debugger
//
//  "AWT-Callback"   enqueue events reported by window manager
//                   paint windows
//
//  "Debugger"       dequeue events reported by remote debugger and window manager
//                   interrogate remote debugger state
//                   issue repaint requests
//                   issue remote debugger commands
//
package jd;
import  java.util.*;
import  java.util.zip.*;
import  java.io.*;
import  sun.tools.debug.*;
import  util.Tracer;
import  ui.FlyoverWindow;
import  ui.MultiLabel;
import  java.awt.Component;

class Debugger extends Thread 
   implements Constants, DebuggerCallback
   {
   static final String         GREETING = "Select a \"main\" class and click \"Release all threads\" button to begin execution";
   
   private String              classPath;               // places to look for .class files
   private String              sourcePath;              // places to look for .java files
   private Vector              classRepositories;       // parsed & validated
   private Vector              sourceRepositories;      // ...versions of above
   
   private boolean             attachToRemoteVM;        // how to debug: remotely vs. locally
   private boolean             initialThreadCreated;    // has debugee thread been created yet?
   private String              initialClassName;        // if so, here's the class whose "main" method was invoked
   
   // when debugging remotely...
   //
   private String              remoteVMHost;            // hostname to connect to
   private String              remoteVMPassword;        // password printed by "java_g -debug" when host was started
   
   // when debugging locally...
   //
   private String              vmStartupArgs;           // arguments with which to start VM to be debugged
   private String              commandLineArgs[];       // arguments with which to start thread to be debugged (null --> choose from class list)
   
   private ClassTree           classTree;               // class repository displayer
   private ClassFinder         classFinder;             // class repository searcher
   private SourceFinder        sourceFinder;            // source repository searcher
   private Vector              breakpointList;          // currently set breakpoints (element type == Breakpoint)
   private String              searchString = "";       // most recent string used for text search
   private String              lineNumberString = "";   // most recent string used for text goto

   // debug agent communication area
   //
   private RemoteDebugger      remoteDebugger;          // connection to remote debugger (null --> not currently connected)

   // display areas and controls
   //
   private DebuggerPanel       debuggerPanel;           // main panel, enclosing the following subpanels...
   private TextPanel           classTreePanel;          // classes to choose for browsing or breakpointing
   private TextPanel           threadsPanel;            // threads running in remote machine
   private TextPanel           framesPanel;             // stackframes of currently selected thread
   private TextPanel           localsPanel;             // local variables of currently selected stackframe
   private TextPanel           valuePanel;              // a variable to be examined in more detail
   private ConsolePanel        consolePanel;            // record of stdin, stdout, stderr
   private TextPanel           sourcePanel;             // source of currently selected stackframe or class
   private MessagePanel        messagePanel;            // messages
   private ButtonPanel         buttonPanel;             // control buttons

   // current context
   //
   private RemoteThread        currentThreadSelection;  // currently selected thread (eg. for stepping)
   private RemoteClass         currentClassSelection;   // currently selected class (eg. for breakpointing)
   
   // snapshots to feed preceeding panels
   //
   private ClassTreeSnapshot   classTreeSnapshot;
   private ThreadsSnapshot     threadsSnapshot; 
   private FramesSnapshot      framesSnapshot;
   private LocalsSnapshot      localsSnapshot;
   private ValueSnapshot       valueSnapshot;
   private SourceSnapshot      sourceSnapshot;
   
   // Create main panel.
   //
   Debugger(String classPath, String sourcePath, boolean attachToRemoteVM, String remoteVMHost, String remoteVMPassword, String vmStartupArgs, String commandLineArgs[])
      {
      super("Debugger");
      this.classPath          = classPath;
      this.sourcePath         = sourcePath;
      this.classRepositories  = new Vector();
      this.sourceRepositories = new Vector();
      this.attachToRemoteVM   = attachToRemoteVM;
      this.remoteVMHost       = remoteVMHost;
      this.remoteVMPassword   = remoteVMPassword;
      this.vmStartupArgs      = vmStartupArgs;
      this.commandLineArgs    = commandLineArgs;
      
      DebuggerEventQueue queue = new DebuggerEventQueue();
      DebuggerFrame      frame = new DebuggerFrame(this, attachToRemoteVM ? "remote client" : commandLineArgs != null ? commandLineArgs[0] : "");
      
      Session.init(queue, frame);
      
      debuggerPanel  = frame.debuggerPanel;
      classTreePanel = debuggerPanel.classesPanel;
      threadsPanel   = debuggerPanel.threadsPanel;
      framesPanel    = debuggerPanel.framesPanel;
      localsPanel    = debuggerPanel.localsPanel;
      valuePanel     = debuggerPanel.inspectorPanel;
      consolePanel   = debuggerPanel.consolePanel;
      sourcePanel    = debuggerPanel.sourcePanel;
      messagePanel   = debuggerPanel.messagePanel;
      buttonPanel    = debuggerPanel.buttonPanel;
      }
      
   // Run "Debugger" thread.
   //
   public void
   run()
      {
      startBlinking();
    
      // give AWT thread time to start up and perform initial rendering, so we can show "loading" progress
      try { Thread.currentThread().sleep(1000); } catch (Exception e) {}

      for (boolean firstTime = true;;)
         {
         try
            {
            if (firstTime)
               {
               firstTime          = false;
               classRepositories  = parseAndValidatePathSpecification("classpath",  classPath,  CLASS_PATH_SEPARATOR );
               sourceRepositories = parseAndValidatePathSpecification("sourcepath", sourcePath, SOURCE_PATH_SEPARATOR);
               classTree          = new ClassTree(classRepositories);
               classFinder        = new ClassFinder(classRepositories);
               sourceFinder       = new SourceFinder(sourceRepositories);
               breakpointList     = new Vector();
               
               doConnect();
               buttonPanel.enableAllButtons();
               if (attachToRemoteVM)
                  ;
               else if (commandLineArgs != null)
                  doCreateInitialThread(commandLineArgs);
               else
                  {
                  inform(GREETING);
                  stopBlinking();
                  }
               doSnapshot(true);
               continue;
               }

            Session.frame.setNormalCursor();
            DebuggerEvent ev = Session.eventQueue.get();
            Session.frame.setBusyCursor();

            switch (ev.kind)
               {
               case DebuggerEvent.BUTTON_SOURCEPATH:
               {
               inform("");
               PathDialog dialog = new PathDialog(Session.frame, "Jd source path", sourcePath, SOURCE_PATH_SEPARATOR);
               String     result = dialog.edit();
               if (result != null)
                  {
                  sourcePath         = result;
                  sourceRepositories = parseAndValidatePathSpecification("sourcepath", sourcePath, SOURCE_PATH_SEPARATOR);
                  sourceFinder       = new SourceFinder(sourceRepositories);
                  sourceSnapshot.invalidate(); // force reload from filesystem, even if file appears to be "the same"
                  sourceSnapshot = new SourceSnapshot(sourceFinder, currentClassSelection, breakpointList);
                  sourcePanel.synchronizeView();
                  }
               break;
               }
               
               case DebuggerEvent.BUTTON_CLASSPATH:
               {
               inform("");
               PathDialog dialog = new PathDialog(Session.frame, "Jd class path", classPath, CLASS_PATH_SEPARATOR);
               String     result = dialog.edit();
               if (result != null)
                  {
                  classPath         = result;
                  classRepositories = parseAndValidatePathSpecification("classpath", classPath, CLASS_PATH_SEPARATOR);
                  classTree         = new ClassTree(classRepositories);
                  classFinder       = new ClassFinder(classRepositories);
      
                  startBlinking();
                  doRestart();
                  inform(GREETING);
                  doSnapshot(true);
                  }
               break;
               }
               
               case DebuggerEvent.BUTTON_SHUTDOWN:
               // Note: if we simply call exit(), then this process will block
               // on the socket to the remote debugger (at least on Windows95)
               // and we won't terminate until the remote debugger is killed manually.
               // The call here to disconnect() closes our side of the socket and allows
               // us to terminate cleanly. Closing our socket also causes the remote
               // debugger to terminate cleanly.
               //
               buttonPanel.disableAllButtons();
               doClean();
               doDisconnect();
               
               // save our state so we can restore it in subsequent reincarnations
               //
               debuggerPanel.putProperties();
               Session.frame.putProperties();
               Session.saveProperties();
               System.exit(0);
               break;

               case DebuggerEvent.BUTTON_RESTART:
               startBlinking();
               doRestart();
               if (commandLineArgs == null && initialClassName != null)
                  {
                  classTreeSnapshot = new ClassTreeSnapshot(classTree, currentClassSelection = remoteDebugger.findClass(initialClassName));
                  classTreePanel.synchronizeView();
                  sourceSnapshot = new SourceSnapshot(sourceFinder, currentClassSelection, breakpointList);
                  sourcePanel.synchronizeView();
                  }
               else
                  doSnapshot(true);
               break;
               
               case DebuggerEvent.BUTTON_STOP:
               stopBlinking();
               doStop(true);
               doSnapshot(true);
               break;

               case DebuggerEvent.BUTTON_GO:
               startBlinking();
               if (initialThreadCreated == false && attachToRemoteVM == false)
                  { // launch a new thread
                  if (commandLineArgs == null)
                     { // launch from list
                     if (currentClassSelection == null)
                        {
                        informAndBeep("Please select a class to run");
                        break;
                        }
                     String args[] = new String[1];
                     args[0] = currentClassSelection.getName();
                     doCreateInitialThread(args);
                     }
                  else
                     { // launch from command line
                     doCreateInitialThread(commandLineArgs);
                     }
                  }
               else
                  { // resume execution of existing thread(s)
                  currentThreadSelection = null;                            // accept breakpoints from any thread
                  doGo(true);
                  }
               break;

               case DebuggerEvent.BUTTON_RUN:
               startBlinking();
               currentThreadSelection = threadsSnapshot.currentSelection(); // accept breakpoints from this thread only
               doGo(true);
               break;
               
               case DebuggerEvent.BUTTON_STEP_OVER:
               startBlinking();
               currentThreadSelection = threadsSnapshot.currentSelection(); // accept breakpoints from this thread only
               doStepOver();
               break;

               case DebuggerEvent.BUTTON_STEP_IN:
               startBlinking();
               currentThreadSelection = threadsSnapshot.currentSelection(); // accept breakpoints from this thread only
               doStepIn();
               break;

               case DebuggerEvent.BUTTON_STEP_OUT:
               startBlinking();
               currentThreadSelection = threadsSnapshot.currentSelection(); // accept breakpoints from this thread only
               doStepOut();
               break;

               case DebuggerEvent.BUTTON_SNAPSHOT:
               inform("");
               doSnapshot(false);
               break;

               case DebuggerEvent.BUTTON_EDIT:
               inform("");
               if (sourceSnapshot != null)
                  sourceSnapshot.edit(sourceFinder);
               break;

               case DebuggerEvent.PICK:
               {
               inform("");
               Snapshot target = (Snapshot)ev.target;
               int      lino   = ev.lino;
                    if (target instanceof ClassTreeSnapshot) doPickClassTree((ClassTreeSnapshot)target,lino);
               else if (target instanceof ThreadsSnapshot)   doPickThread((ThreadsSnapshot)target,     lino);
               else if (target instanceof FramesSnapshot)    doPickFrame((FramesSnapshot)target,       lino);
               else if (target instanceof LocalsSnapshot)    doPickLocal((LocalsSnapshot)target,       lino);
               else if (target instanceof ValueSnapshot)     doPickValue((ValueSnapshot)target,        lino);
               else if (target instanceof SourceSnapshot)    doPickSourceLine((SourceSnapshot)target,  lino);
               break;
               }
               
               case DebuggerEvent.SEARCHF:
               case DebuggerEvent.SEARCHB:
               {
               StringDialog dialog = new StringDialog(Session.frame, "Enter Search String", searchString);
               String       result = dialog.edit();
               if (result != null)
                  {
                  searchString = result;
                  ((TextPanel)ev.target).search(searchString, ev.kind == DebuggerEvent.SEARCHF ? +1 : -1);
                  }
               break;
               }

               case DebuggerEvent.GOTO:
               {
               StringDialog dialog = new StringDialog(Session.frame, "Enter Line Number", lineNumberString);
               String       result = dialog.edit();
               if (result != null)
                  {
                  int lino = 0;
                  try { lino = Integer.valueOf(result).intValue(); } catch (NumberFormatException e) {}
                  if (lino > 0)
                     {
                     lineNumberString = result;
                     ((TextPanel)ev.target).gotoLine(lino - 1);
                     }
                  }
               break;
               }

               case DebuggerEvent.ZOOM:
               debuggerPanel.viewPanel.zoom((Component)ev.target);
               break;

               case DebuggerEvent.REMOTE_BREAKPOINT:
               {
               RemoteThread thread  = (RemoteThread)ev.target;
               String       message = isAtBreakpoint(thread) ? "Thread arrived at breakpoint" : "Thread finished stepping";
               if (currentThreadSelection != null && currentThreadSelection != thread)
                  { // some thread that we're not interested in right now
               // System.out.println(message + ": " + thread.getName());
                  doGo(false);
                  continue;
                  }
                  
               stopBlinking();
               currentThreadSelection = thread;
               currentClassSelection  = null;
               inform(message);
               doSnapshot(true);
               break;
               }

               case DebuggerEvent.REMOTE_EXCEPTION:
               {
               RemoteThread thread  = (RemoteThread)ev.target;
               String       message = ev.text;
               //
               // The message is something like "Uncaught exception: MMM\t at XXX(YYY:NNN)\n
               //                                   at XXX(YYY:NNN)\n
               //                                   at XXX(YYY:NNN)\n
               //                                   ..."
               // The "Uncaught" part of the message is misleading - if we resume() the thread,
               // somebody might still catch the exception and execution can still proceed.
               //
               // We'll just display the "MMM" part of the message and ignore the rest.
               //
               message = "Thread about to throw an exception: " + message.substring(message.indexOf(": ") + 2, message.indexOf("\t"));
               
               if (currentThreadSelection != null && currentThreadSelection != thread)
                  { // some thread that we're not interested in right now
               // System.out.println(message + ": " + thread.getName());
                  doGo(false);
                  continue;
                  }
                  
               stopBlinking();
               currentThreadSelection = thread;
               currentClassSelection  = null;
               inform(message);
               doSnapshot(true);
               break;
               }

               case DebuggerEvent.REMOTE_THREAD_DEATH:
               {
               RemoteThread thread  = (RemoteThread)ev.target;
               String       message = "A thread died";
               
               if (currentThreadSelection != null && currentThreadSelection != thread)
                  { // some thread that we're not interested in right now
               // System.out.println(message + ": " + thread.getName());
                  doGo(false);
                  continue;
                  }
                  
               stopBlinking();
               currentThreadSelection = thread;
               currentClassSelection  = null;
               inform(message);
               doSnapshot(true);
               break;
               }

               case DebuggerEvent.REMOTE_QUIT:
               stopBlinking();
               inform("Remote debugger disconnected");
               buttonPanel.disableAllButtons();
               remoteDebugger = null;
               doClean();
               break;
               }
            }
         
         catch (Exception e)
            {
            for (;;)
               {
               informAndBeep("Internal error: " + e);
               e.printStackTrace();
               if (e instanceof java.net.SocketException && e.getMessage().equals("Connection reset by peer"))
                  { // error on remote side
                  inform("Remote debugger disconnected");
                  buttonPanel.disableAllButtons();
                  remoteDebugger = null;
                  try { doClean(); } catch(Exception x) { e = x; continue; }
                  }
               break;
               }
            }
         }
      }

   // Kill current VM and spawn another one using same breakpoints and startup method.
   //
   synchronized
   private void
   doRestart()
      throws Exception
      {
      // no restart capability when attached to already-running VM
      //
      if (attachToRemoteVM)
         {
         informAndBeep("sorry, no remote vm restart capability");
         return;
         }

      buttonPanel.disableAllButtons();

      // release all remote references
      //
      doClean();

      // close connection to remote virtual machine
      //
      doDisconnect();

      // instantiate a new remote virtual machine
      //
      doConnect();
      buttonPanel.enableAllButtons();

      // restore breakpoints
      //
      for (Enumeration b = breakpointList.elements(); b.hasMoreElements(); )
         {
         Breakpoint bp = (Breakpoint)b.nextElement();

         RemoteClass remoteClass = remoteDebugger.findClass(bp.className);
         if (remoteClass == null)
            {
            Session.popup("restart: can't restore breakpoint in " + bp.className + ": no such class");
            continue;
            }

         String message = remoteClass.setBreakpointLine(bp.lineNumber);
         if (message.length() != 0)
            {
            Session.popup("restart: can't restore breakpoint at line " + bp.lineNumber + " of " + bp.sourceFileName + ": " + message);
            continue;
            }
         }

      // start debugee thread
      //
      if (commandLineArgs != null && attachToRemoteVM == false)
         doCreateInitialThread(commandLineArgs);
      }
      
   // Establish connection to debugee.
   //
   synchronized
   private void
   doConnect()
      throws Exception
      {
      inform("Connecting (this may take a moment)");
      initialThreadCreated = false;
      boolean verbose = false;
      if (attachToRemoteVM)
         remoteDebugger = new RemoteDebugger(remoteVMHost, remoteVMPassword, this, verbose);
      else
         {
         String args = vmStartupArgs + " -classpath " + classPath;
      // System.out.println("jd: remote vm args are " + args);
         remoteDebugger = new RemoteDebugger(args, this, verbose);
         }
      inform("Connected");
      }
         
   // Release connection to debugee.
   //
   synchronized
   private void
   doDisconnect()
      throws Exception
      {
      inform("Disconnecting");
      if (remoteDebugger != null)
         remoteDebugger.close();
      remoteDebugger = null;
      inform("Disconnected");
      }
      
   // Create initial debugee thread.
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doCreateInitialThread(String commandLineArgs[])
      throws Exception
      {
      String clsName    = commandLineArgs[0];
      String methodName = "main";
      String fullName   = clsName + ".main";
      
      inform("Loading " + fullName);

      RemoteClass cls = null;
      try { cls = remoteDebugger.findClass(clsName); } catch (IllegalArgumentException e) { } //!!TODO: unfortunately, the remote debugger never actually throws this exception (ie. for a malformed class name)
      if (cls == null)
         {
         informAndBeep("Can't find " + fullName + ": no such class");
         return;
         }

      RemoteField method = null; 
      try { method = cls.getMethod(methodName); } catch (NoSuchMethodException e) { }
      if (method == null)
         {
         informAndBeep("Can't find " + fullName + ": no such method");
         return;
         }

      // set breakpoint on main entrypoint
      //
      String msg = cls.setBreakpointMethod(method);
      if (msg.length() != 0)
         {
         informAndBeep("Can't breakpoint " + fullName);
         return;
         }

      // bring our breakpoint list into correspondence with that of remote vm
      //
      breakpointList = new Vector();
      String breakpoints[] = remoteDebugger.listBreakpoints();
      for (int i = 0, n = breakpoints.length; i < n; ++i)
         {
         String      breakpoint       = breakpoints[i]; // a string of the form "className:lineNumber"
         int         colonIndex       = breakpoint.indexOf(":");
         String      bpClassName      = breakpoint.substring(0, colonIndex);
         int         bpLineNumber     = Integer.valueOf(breakpoint.substring(colonIndex + 1)).intValue();
         RemoteClass bpClass          = remoteDebugger.findClass(bpClassName);
         String      bpSourceFileName = bpClass.getSourceFileName();

         breakpointList.addElement(new Breakpoint(bpSourceFileName, bpLineNumber, bpClassName));
         }

      // kick off thread
      //
      RemoteThreadGroup group = remoteDebugger.run(commandLineArgs.length, commandLineArgs);
      if (group == null)
         {
         informAndBeep("Can't load " + fullName);
         return;
         }

      initialThreadCreated = true;
      initialClassName     = clsName;
      inform("Thread created and running");
      }
      
   // Cease execution of selected thread and destroy its corresponding thread object.
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doDelete()
      throws Exception
      {
      RemoteThread thread = currentThreadSelection;

      if (!isControllable(thread, true))
         return;

      inform("Thread deleted");
      thread.stop();
      }
      
   // Stop execution of all non-system threads (a stopped thread can still be stepped).
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doStop(boolean inform)
      throws Exception
      {
      int          cnt = 0;
      RemoteThread threads[] = getUserThreads();
      for (int i = 0, n = threads.length; i < n; ++i)
         {
         RemoteThread thread = threads[i];
         if (!thread.isSuspended())
            {
            thread.suspend();
            ++cnt;
            }
         }
      if (inform)
         inform(cnt + " thread(s) stopped");
      }
      
   // Allow stopped threads to resume execution.
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doGo(boolean informAndSynchronizeView)
      throws Exception
      {
      int          cnt = 0;
      RemoteThread threads[] = getUserThreads();
      for (int i = 0, n = threads.length; i < n; ++i)
         {
         RemoteThread thread = threads[i];

    // !! for jdk 1.1.3...
    //   if (thread.isSuspended())
    //      thread.resume();
    //   else
    //      thread.cont();

         // for jdk 1.1.4 it appears that we must issue both "resume" and "cont"
         // in order to handle the case where we've "stepped in" from a breakpointed line
         // and wish now to "resume" execution, even though we're no longer at a breakpoint.
         //
         // if we're not actually at a breakpoint, the "cont" seems to be harmless.
         //
         thread.resume();
         thread.cont();

         ++cnt;
         }
      if (informAndSynchronizeView)
         {
         inform(cnt + " thread(s) released");
         threadsSnapshot = new ThreadsSnapshot(getUserThreads(), currentThreadSelection);
         threadsPanel.synchronizeView();
         }
      }

   // Release selected thread for one line of execution (stepping over method call, if any).
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doStepOver()
      throws Exception
      {
      RemoteThread thread = currentThreadSelection;

      if (!isControllable(thread, true))
         return;

      if (!isSteppable(thread, true))
         return;

      inform("Thread stepping");
      try                          { thread.next(); }
      catch (IllegalAccessError e) { informAndBeep("Can't step: " + e); }
      }
      
   // Release selected thread for one line of execution (stepping into method call, if any).
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doStepIn()
      throws Exception
      {
      RemoteThread thread = currentThreadSelection;

      if (!isControllable(thread, true))
         return;

      if (!isSteppable(thread, true))
         return;

      inform("Thread stepping");
      try                          { thread.step(true); }
      catch (IllegalAccessError e) { informAndBeep("Can't step: " + e); }
      }
      
   // Release selected thread for execution until control returns to caller of current method.
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doStepOut()
      throws Exception
      {
      RemoteThread thread = currentThreadSelection;

      if (!isControllable(thread, true))
         return;

      if (!isSteppable(thread, true))
         return;

      inform("Thread stepping");
      try                          { thread.stepOut(); }
      catch (IllegalAccessError e) { informAndBeep("Can't step: " + e); }
      }
      
   // Take a snapshot of the remote machine and update all panels.
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doSnapshot(boolean chooseInnermostFrame)
      throws Exception
      {
      if (chooseInnermostFrame)
         framesSnapshot = null;
      
      threadsSnapshot = new ThreadsSnapshot(getUserThreads(), currentThreadSelection);
      threadsPanel.synchronizeView();
      
      framesSnapshot = new FramesSnapshot(threadsSnapshot.currentSelection(), framesSnapshot == null ? null : framesSnapshot.currentSelection());
      framesPanel.synchronizeView();
      
      localsSnapshot = new LocalsSnapshot(framesSnapshot.currentSelection(), localsSnapshot == null ? null : localsSnapshot.currentSelection());
      localsPanel.synchronizeView();

      valueSnapshot = new ValueSnapshot(localsSnapshot.currentSelection());
      valuePanel.synchronizeView();

      sourceSnapshot = new SourceSnapshot(sourceFinder, framesSnapshot.currentSelection(), breakpointList);
      sourcePanel.synchronizeView();
      
      classTreeSnapshot = new ClassTreeSnapshot(classTree, currentClassSelection = framesSnapshot.currentSelection() == null ? null : framesSnapshot.currentSelection().getRemoteClass());
      classTreePanel.synchronizeView();
      }
   
   // Create an empty snapshot (thereby releasing all local references to remote objects).
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doClean()
      throws Exception
      {
      threadsSnapshot = new ThreadsSnapshot(new RemoteThread[0], null);
      threadsPanel.synchronizeView();
      
      framesSnapshot = new FramesSnapshot(currentThreadSelection = null, null);
      framesPanel.synchronizeView();
      
      localsSnapshot = new LocalsSnapshot((RemoteStackFrame)null, null);
      localsPanel.synchronizeView();

      valueSnapshot = new ValueSnapshot((RemoteStackVariable)null);
      valuePanel.synchronizeView();

      sourceSnapshot = new SourceSnapshot(sourceFinder, (RemoteStackFrame)null, new Vector());
      sourcePanel.synchronizeView();
      
      classTreeSnapshot = new ClassTreeSnapshot(classTree, currentClassSelection = null);
      classTreePanel.synchronizeView();
      }

   // Take a snapshot of a particular remote thread and update appropriate panels.
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doPickThread(ThreadsSnapshot oldSnapshot, int lino)
      throws Exception
      {
   // Tracer.trace("pick", "threads");
   
      threadsSnapshot = new ThreadsSnapshot(oldSnapshot, lino);
      threadsPanel.synchronizeView();
      
      framesSnapshot = new FramesSnapshot(currentThreadSelection = threadsSnapshot.currentSelection(), null);
      framesPanel.synchronizeView();
      
      localsSnapshot = new LocalsSnapshot(framesSnapshot.currentSelection(), null);
      localsPanel.synchronizeView();
      
      valueSnapshot = new ValueSnapshot(localsSnapshot.currentSelection());
      valuePanel.synchronizeView();
      
      sourceSnapshot = new SourceSnapshot(sourceFinder, framesSnapshot.currentSelection(), breakpointList);
      sourcePanel.synchronizeView();
      
      classTreeSnapshot = new ClassTreeSnapshot(classTree, currentClassSelection = framesSnapshot.currentSelection() == null ? null : framesSnapshot.currentSelection().getRemoteClass());
      classTreePanel.synchronizeView();
      }

   // Take a snapshot of a particular remote frame and update appropriate panels.
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doPickFrame(FramesSnapshot oldSnapshot, int lino)
      throws Exception
      {
   // Tracer.trace("pick", "frame");
   
      framesSnapshot = new FramesSnapshot(oldSnapshot, lino);
      framesPanel.synchronizeView();
      
      localsSnapshot = new LocalsSnapshot(framesSnapshot.currentSelection(), null);
      localsPanel.synchronizeView();
      
      valueSnapshot = new ValueSnapshot(localsSnapshot.currentSelection());
      valuePanel.synchronizeView();
      
      sourceSnapshot = new SourceSnapshot(sourceFinder, framesSnapshot.currentSelection(), breakpointList);
      sourcePanel.synchronizeView();
      
      classTreeSnapshot = new ClassTreeSnapshot(classTree, currentClassSelection = framesSnapshot.currentSelection() == null ? null : framesSnapshot.currentSelection().getRemoteClass());
      classTreePanel.synchronizeView();
      }
   
   // Take a snapshot of a particular local variable and update appropriate panels.
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doPickLocal(LocalsSnapshot oldSnapshot, int lino)
      throws Exception
      {
   // Tracer.trace("pick", "local");
   
      localsSnapshot = new LocalsSnapshot(oldSnapshot, lino);
      localsPanel.synchronizeView();
      
      valueSnapshot = new ValueSnapshot(localsSnapshot.currentSelection());
      valuePanel.synchronizeView();
      }
   
   // Expand detail of a "watched value" and update appropriate panels.
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doPickValue(ValueSnapshot oldSnapshot, int lino)
      throws Exception
      {
   // Tracer.trace("pick", "value");
   
      valueSnapshot = new ValueSnapshot(oldSnapshot, lino);
      valuePanel.synchronizeView();
      }
   
   // Choose a class for browsing or breakpointing.
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doPickClassTree(ClassTreeSnapshot oldSnapshot, int lino)
      throws Exception
      {
   // Tracer.trace("pick", "classtree");
   
      classTreeSnapshot = new ClassTreeSnapshot(oldSnapshot, lino);
      classTreePanel.synchronizeView();
   
      String className = classTreeSnapshot.currentSelection();
      if (className == null)
         { // toggled tree visibility, didn't click on a class
         return;
         }

      sourceSnapshot = new SourceSnapshot(sourceFinder, currentClassSelection = remoteDebugger.findClass(className), breakpointList);
      sourcePanel.synchronizeView();
      }
   
   // Take a snapshot of a particular remote source file and toggle a breakpoint.
   // Runs in "Debugger" thread.
   //
   synchronized
   private void
   doPickSourceLine(SourceSnapshot oldSnapshot, int lino)
      throws Exception
      {
   // Tracer.trace("pick", "source");
   
      RemoteClass target = currentClassSelection;
      if (target == null)
         {
         informAndBeep("No breakpoint target available");
         return;
         }

      int     lineNumber = lino + 1;
      boolean clearbp    = sourceSnapshot.hasBreakpointAt(new Integer(lino));
      String  message    = clearbp ? target.clearBreakpointLine(lineNumber) : target.setBreakpointLine(lineNumber);

      if (message.startsWith("No code at"))
         {
         // The debug api references breakpoints by class name, not by source file name (even though
         // breakpoint line numbers are file-relative). Thus if multiple classes (anonymous or inner classes,
         // for example) were produced from the same source file, it's pretty tough to figure out which class 
         // the user has clicked on. Our only recourse is to breakpoint every class that came from the source file. 
         // This is a royal hack.
         
         inform("Figuring out classes comprising " + target.getSourceFileName() + "...");
         
         Vector classNames  = classFinder.classesForSource(sourceFinder.resolveSourceFileName(sourceSnapshot.packageName, sourceSnapshot.remoteSourceName));
         for (Enumeration c = classNames.elements(); c.hasMoreElements(); )
            {
            String className = (String)c.nextElement();
            inform("Trying class " + className + " as breakpoint target");
            target = remoteDebugger.findClass(className);
            if (target == null)
               continue;
            message = clearbp ? target.clearBreakpointLine(lineNumber) : target.setBreakpointLine(lineNumber);
            if (message.length() == 0)
               break;
            }
         }
            
      if (message.length() != 0)
         {
         informAndBeep(message);
         return;
         }
      
      String sourceFileName = target.getSourceFileName();
      String className      = target.getName();
      
      if (clearbp)
         {
         for (int i = 0; i < breakpointList.size(); ++i)
            {
            Breakpoint bp = (Breakpoint)breakpointList.elementAt(i);
            if (bp.sourceFileName.equals(sourceFileName) && bp.lineNumber == lineNumber)
               breakpointList.removeElementAt(i);
            }
         inform("Breakpoint removed");
         }
      else
         {
         breakpointList.addElement(new Breakpoint(sourceFileName, lineNumber, className));
         inform("Breakpoint set");
         }

      // update source display with breakpoint icon
      //
      sourceSnapshot = new SourceSnapshot(sourceSnapshot, breakpointList);
      sourcePanel.synchronizeView();

      // match class display to breakpoint
      //
      classTreeSnapshot = new ClassTreeSnapshot(classTree, currentClassSelection = target);
      classTreePanel.synchronizeView();

      // update thread display with "at breakpoint" status
      //
      threadsSnapshot = new ThreadsSnapshot(getUserThreads(), currentThreadSelection);
      threadsPanel.synchronizeView();
      }
   
   // Is specified thread ok to stop/go/step/kill?
   // Runs in "Debugger" thread.
   //
   synchronized
   private boolean
   isControllable(RemoteThread thread, boolean complain)
      throws Exception
      {
      if (thread == null)
         {
         if (complain) 
            informAndBeep("Please select a thread");
         return false;
         }

      return true;
      }

   // Is specified thread currently in a state in which its execution can be stepped forward?
   // Runs in "Debugger" thread.
   //
   synchronized
   private boolean
   isSteppable(RemoteThread thread, boolean complain)
      throws Exception
      {
      if (isAtBreakpoint(thread) || thread.getStatus().equals("suspended"))
         return true;

      if (complain)
         informAndBeep("Can't step: thread is already running");

      return false;
      }

   // Is specified thread currently at a breakpoint?
   // Runs in "Debugger" thread.
   //
   synchronized
   private boolean
   isAtBreakpoint(RemoteThread thread)
      throws Exception
      {
      if (thread == null)
         { // jjb says thread is sometimes null, but I don't understand why
         return false;
         }
      return thread.getStatus().equals("at breakpoint");
      }

   // Is a subpanel currently in "zoomed" state?
   //
   boolean
   isZoomed()
      {
      return debuggerPanel.viewPanel.zoomParent != null;
      }

   // Get all user threads currently running in the remote jvm.
   //
   synchronized
   private RemoteThread[]
   getUserThreads()
      throws Exception
      {
      RemoteThreadGroup root          = remoteDebugger.listThreadGroups(null)[0];
      RemoteThread      allThreads[]  = root.listThreads(true);
      RemoteThread      userThreads[] = new RemoteThread[allThreads.length];
      int               userCnt       = 0;

      for (int i = 0, n = allThreads.length; i < n; ++i)
         {
         RemoteThread      thread = allThreads[i];
         RemoteThreadGroup group  = (RemoteThreadGroup)thread.getFieldValue("group");
         if (group != null)
            { // jjb says group is sometimes null, but I don't understand why
            String name = group.getName();

            if (name.equals("system"))
               continue; // threads for finalizer, awt, debug agent, etc.
            
            if (name.equals("main") && attachToRemoteVM == false)
               continue; // dummy thread, seems to do nothing
            }
         userThreads[userCnt++] = thread;
         }

      RemoteThread result[] = new RemoteThread[userCnt];
      for (int i = 0; i < userCnt; ++i)
         result[i] = userThreads[i];
      
      return result;
      }

   // Fetch latest snapshot to be displayed in a panel.
   // Runs in "Debugger" thread.
   //
   synchronized
   Snapshot
   getSnapshotFor(TextPanel panel)
      {
      if (panel == classTreePanel) return classTreeSnapshot;
      if (panel == threadsPanel)   return threadsSnapshot;
      if (panel == framesPanel)    return framesSnapshot;
      if (panel == localsPanel)    return localsSnapshot;
      if (panel == valuePanel)     return valueSnapshot;
      if (panel == sourcePanel)    return sourceSnapshot;
      return null;
      }

   // Display message.
   //
   synchronized
   void
   inform(String message)
      {
   // if (message.length() != 0) Tracer.trace("=== " + message + " ===");
      messagePanel.setMessage(message);
      }

   // Display message, with a beep.
   //
   synchronized
   void
   informAndBeep(String message)
      {
      inform(message);
      java.awt.Toolkit.getDefaultToolkit().beep();
      }

   void
   startBlinking()
      {
      buttonPanel.snapshotButton.startBlinking(DebuggerFrame.refreshButtonAltImage);
      }

   void
   stopBlinking()
      {
      buttonPanel.snapshotButton.stopBlinking();
      }

   // Parse "pathSpecification" and return names of path components that actually exist.
   // A path component may be a filesystem directory, a .zip file, or a .jar file.
   //
   static Vector
   parseAndValidatePathSpecification(String pathKind, String pathSpecification, char pathSeparator)
      {
      Vector components = new Vector();
      for (StringTokenizer st = new StringTokenizer(pathSpecification, String.valueOf(pathSeparator), false); st.hasMoreTokens(); )
         {
         String componentName = st.nextToken();
         
         boolean isDuplicate = false;
         for (Enumeration c = components.elements(); c.hasMoreElements(); )
            if (componentName.equals((String)c.nextElement()))
               {
               isDuplicate = true;
               break;
               }
         
         if (isDuplicate)
            {
            Session.warn(pathKind + " component \"" + componentName + "\" multiply defined (duplicate ignored)");
            continue;
            }
         
         if (componentName.endsWith(".zip") || componentName.endsWith(".jar"))
            {
            try
               {
               ZipFile archive = new ZipFile(componentName);
               archive.close();
               }
            catch(IOException x) // should be FileNotFoundException not IOException - Sun got it wrong
               {
               Session.warn(pathKind + " component \"" + componentName + "\" doesn't exist (ignoring it)");
               continue;
               }
            }
         else
            {
            File dir = new File(componentName);

            if (dir.exists() == false)
               {
               Session.warn(pathKind + " component \"" + componentName + "\" doesn't exist (ignoring it)");
               continue;
               }

            if (dir.isDirectory() == false)
               {
               Session.warn(pathKind + " component \"" + componentName + "\" is not a directory (ignoring it)");
               continue;
               }
            }
         
         components.addElement(componentName);
         }
      
      return components;
      }

   //-------------------------------------//
   // Remote debugger callback interface. //
   //-------------------------------------//
   
   // Called when remote machine wants to read from console.
   //
   public char
   readFromConsole() //- implements DebuggerCallback
      { 
      return consolePanel.get();
      }

   // Called when remote machine wants to write to console.
   //
   public void 
   printToConsole(String msg) //- implements DebuggerCallback
      { 
      consolePanel.put(msg);
      }

   // Called when a remote thread stops at breakpoint or completes line-, call-, or instruction- step.
   //
   public void 
   breakpointEvent(RemoteThread remoteThread) //- implements DebuggerCallback
      {
      Session.eventQueue.put(new DebuggerEvent(DebuggerEvent.REMOTE_BREAKPOINT, remoteThread));
      }
      
   // Called when a remote thread throws an exception.
   //
   public void 
   exceptionEvent(RemoteThread remoteThread, String message) //- implements DebuggerCallback
      {
      Session.eventQueue.put(new DebuggerEvent(DebuggerEvent.REMOTE_EXCEPTION, remoteThread, message));
      }

   // Called when a remote thread finishes processing a ThreadDeath exception (ie. after it has died).
   //
   public void 
   threadDeathEvent(RemoteThread remoteThread) //- implements DebuggerCallback
      {
      Session.eventQueue.put(new DebuggerEvent(DebuggerEvent.REMOTE_THREAD_DEATH, remoteThread));
      }
      
   // Called when connection to remote debugger is lost.
   //
   public void quitEvent() //- implements DebuggerCallback
      {
      Session.eventQueue.put(new DebuggerEvent(DebuggerEvent.REMOTE_QUIT));
      }
   }
