// Replacement for AWT `TextArea' class, with additional functionality.
// 09 Sep 1997 Derek Lieber
//
// Note: all line and column coordinates are "0-indexed"
//
package ui;
import  java.awt.*;
import  java.awt.event.*;
import  java.io.*;
import  java.util.Vector;

// A mouse-sensitive text viewport.
// If the text doesn't fit the window bounds, scrollbars can be supplied.
// The keyboard's cursor keys can be used for line and page scrolling.
// Images can be applied to decorate individual lines, at the left edge.
//
public class TextScroller extends FastContainer
   implements Constants, KeyListener, TextScrollbarListener
   {
                                              // edging styles...
   public static final int NO_EDGING     = 0; // ...no edging
   public static final int INSET_EDGING  = 1; // ...shadowed edging, "inset" look
   public static final int OUTSET_EDGING = 2; // ...shadowed edging, "outset" look
   
   public  TextCanvas    canvas;      // text display area (public, to allow listening for mouse motions to implement bug workarounds in AWT 1.1.4 enter/exit reports)
   private TextScrollbar vbar;        // vertical scrollbar
   private TextScrollbar hbar;        // horizontal scrollbar

   private int           searchLino;  // line at which to begin/resume string searches
   private int           edgingStyle; // edging to apply (one of the edging styles above)
   private boolean       barsOk;      // supply scrollbars if needed?
   private boolean       needsRedraw; // has internal state changed?

   //-----------//
   // interface //
   //-----------//
   
   // Create.
   //
   public 
   TextScroller()
      {
      canvas       = new TextCanvas();
      vbar         = new TextScrollbar(true);
      hbar         = new TextScrollbar(false);
      searchLino   = 0;
      edgingStyle  = INSET_EDGING;
      barsOk       = true;
      needsRedraw  = true;

      setLayout(null);
      add(canvas);
      add(vbar);
      add(hbar);

      canvas.addKeyListener(this);
      vbar.addListener(this);
      hbar.addListener(this);

   // setCursor(STANDARD_CURSOR);
   // canvas.setCursor(STANDARD_CURSOR);
   // vbar.setCursor(STANDARD_CURSOR);
   // hbar.setCursor(STANDARD_CURSOR);
      }

   // Listen for clicks on text area.
   //
   synchronized
   public void
   addTextListener(TextListener listener)
      {
      canvas.addTextListener(listener);
      }

   // Specify edging style.
   //
   synchronized
   public void
   setEdging(int edgingStyle)
      {
      this.edgingStyle = edgingStyle;
      }

   // Specify text to be displayed.
   //
   synchronized
   public void
   setText(String text)
      {
      searchLino = 0;
      canvas.setText(text); // install new text
      doLayout();           // add/remove/adjust scrollbars to accommodate new text
      }

   // Specify whether or not scrollbars should be provided when text area is too big to fit window.
   //
   synchronized
   public void
   setScrollbarAvailability(boolean b)
      {
      if (barsOk != b)
         {
         barsOk = b;
         
         boolean scrollbarsWereInUse = (vbar.getSize().width * vbar.getSize().height) != 0 ||
                                       (hbar.getSize().width * hbar.getSize().height) != 0 ;
    
         // add/remove scrollbars
         //
         doLayout();            
         
         // decide if text needs to be redrawn
         //
         if (barsOk)
            { // scrollbars possibly added: no text redraw required because doLayout() did it via setValue()
            }
         else if (scrollbarsWereInUse)
            { // scrollbars removed: redraw required to display text formerly obscured by them
            needsRedraw = true;
            }
         }
      }

   // Specify line to bring into view if it's not currently visible.
   //
   synchronized
   public void
   setViewpoint(int lino)
      {
      if (lino == -1)
         return; // don't care about viewpoint
         
      int newtop = canvas.getViewpointAdjustmentValue(lino);
      if (newtop == -1)
         return; // line already in view
         
      vbar.setValue(newtop);
      }

   // Specify starting line for string searches.
   //
   synchronized
   public void
   setSearchLine(int lino)
      {
      searchLino = lino;
      }

   // Specify line to appear at top of view.
   //
   synchronized
   public void
   setTopmostLine(int lino)
      {
      if (lino == -1)
         return; // don't care about viewpoint

      vbar.setValue(lino);
      }

   // Query line appearing at top of view.
   //
   synchronized
   public int
   getTopmostLine()
      {
   // return (int)(vbar.getValue() + 0.5);
      return double2int(vbar.getValue() + 0.5);
      }

   // Specify line to be hilighted (-1 --> none).
   //
   synchronized
   public void
   setMarkedLine(int lino)
      {
      canvas.setMarkedLine(lino);
      }

   // Specify line to be decorated.
   //
   synchronized
   public void
   setDecoratedLine(int lino, Image decoration)
      {
      canvas.setDecoratedLine(lino, decoration);
      }

   // Remove all decorations.
   //
   synchronized
   public void
   undecorate()
      {
      canvas.undecorate();
      }

   // Find string by scanning from current search position.
   // Taken:    string sought
   //           direction to search (+1 --> forward, -1 --> backward)
   // Returned: lino at which string was found (-1 --> not found)
   //
   synchronized
   public int
   search(String sought, int direction)
      {
      return canvas.search(searchLino, sought, direction);
      }
   
   // Has internal state changed?
   //
   synchronized
   public boolean
   needsRedraw()
      {
      boolean rc = needsRedraw | canvas.needsRedraw;
      needsRedraw = canvas.needsRedraw = false;
      return rc;
      }

   // Convert bytes of text into form suitable for display in TextScroller.
   // !!TODO: accept final line even if it's not terminated with \n
   //         so we can call setText() with simple string
   //
   public static String
   stringFromBytes(byte data[])
      {
      // scan, removing \r's and counting \t's
      int tabs = 0;
      for (int i = 0; i < data.length; ++i)
         {
         byte ch = data[i];
         if (ch == '\r')
            data[i] = ASCII_SPACE;
         else if (ch == '\t')
            ++tabs;
         }
      
      if (tabs == 0)
         return new String(data, 0, data.length);
      
      // rescan, expanding \t's
      byte buf[] = new byte[data.length + tabs * 7];
      int  out = 0;
      for (int in = 0, col = 0; in < data.length;)
         {
         byte ch = data[in++];
         buf[out++] = ch;
         if (ch == '\n')
            {
            col = 0;
            continue;
            }
         ++col;
         if (ch == '\t')
            {
            buf[out - 1] = ASCII_SPACE;
            for (; (col % 8) != 0; ++col)
               buf[out++] = ASCII_SPACE;
            }
         }

      return new String(buf, 0, out);
      }

   //------------------------//
   // implementation details //
   //------------------------//
   
   synchronized
   public void
   setName(String name) //- extends Component
      {
      super.setName(name + "Scroller");
      canvas.setName(name + "Canvas");
      }

   synchronized
   public void
   setFont(Font font) //- extends Component
      {
      super.setFont(font);
      canvas.setFont(font);
      }

   synchronized
   public void
   setBackground(Color color) //- extends Component
      {
      super.setBackground(color);
      canvas.setBackground(color);
      }

   synchronized
   public Dimension
   getPreferredSize() //- overrides Component
      {
      return new Dimension(canvas.columns * canvas.fontW + 2 * OUTER_BORDER_THICKNESS,
                           canvas.lines   * canvas.fontH + 2 * OUTER_BORDER_THICKNESS + canvas.fontDescent);
      }
      
   // Accomodate change to window size or amount of text to be displayed.
   //
   synchronized
   public void
   doLayout() //- overrides Container
      {

      // desired layout
      //
      //  +------------------+--+
      //  |                  |  |
      //  |                  |  |
      //  |                  |  |
      //  |                  |  |
      //  |                  |  |
      //  |                  |  |
      //  |                  |  |
      //  +------------------+--+
      //  |                  | *<--- !!TODO: draw window95-style "resize tab" icon here (if both scrollbars are present)
      //  +------------------+

      int x              = OUTER_BORDER_THICKNESS;
      int y              = OUTER_BORDER_THICKNESS;
      int width          = getSize().width  - 2 * OUTER_BORDER_THICKNESS;
      int height         = getSize().height - 2 * OUTER_BORDER_THICKNESS;

      int canvasWidth    = width;
      int canvasHeight   = height;

      int visibleColumns = canvasWidth  / canvas.fontW;
      int visibleLines   = canvasHeight / canvas.fontH;

      // decide which scrollbars (if any) are needed
      //
      int vbarWidth     = 0;
      int hbarHeight    = 0;

      if (barsOk)
         {
         if (canvas.columns > visibleColumns)
            { // horizontal scrollbar needed
            hbarHeight    = SCROLLBAR_THICKNESS;
            canvasHeight -= hbarHeight;
            visibleLines  = canvasHeight / canvas.fontH;
            }

         if (canvas.lines > visibleLines)
            { // vertical scrollbar needed
            vbarWidth      = SCROLLBAR_THICKNESS;
            canvasWidth   -= vbarWidth;
            visibleColumns = canvasWidth  / canvas.fontW;
            }
         }

      // lay out components
      //
      canvas.setBounds(x + 0, y + 0, canvasWidth, canvasHeight);
      vbar.setBounds(x + width - vbarWidth, y + 0, vbarWidth, canvasHeight);
      hbar.setBounds(x + 0, y + height - hbarHeight, canvasWidth, hbarHeight);

      // set scrollbar range and sensitivity
      //
      if (visibleColumns < 1) visibleColumns = 1;
      if (visibleLines   < 1) visibleLines   = 1;

      int maxColumn = canvas.columns - visibleColumns;
      int maxLine   = canvas.lines   - visibleLines;

      if (maxColumn < 0) maxColumn = 0;
      if (maxLine   < 0) maxLine   = 0;

      hbar.setValues(
                    canvas.leftmostColumn, // current thumb position
                    visibleColumns,        // number of positions spanned by thumb
                    0,                     // first thumb position
                    maxColumn              // last thumb position
                    );
      hbar.setBlockIncrement(visibleColumns);

      vbar.setValues(
                    canvas.topmostLine,    // current thumb position
                    visibleLines,          // number of positions spanned by thumb
                    0,                     // first thumb position
                    maxLine                // last thumb position
                    );
      vbar.setBlockIncrement(visibleLines);

      // snug text to bottom/right if window has grown
      //
      if (canvas.topmostLine + visibleLines > canvas.lines)
         vbar.setValue(canvas.lines - visibleLines);
      if (canvas.leftmostColumn + visibleColumns > canvas.columns)
         hbar.setValue(canvas.columns - visibleColumns);
      }

   synchronized
   public void 
   keyPressed(KeyEvent e)  //- implements KeyListener
      {
      switch (e.getKeyCode())
         {
         case KeyEvent.VK_HOME:
         hbar.setValue(0);
         break;

         case KeyEvent.VK_END:
         hbar.setValue(999999);
         break;

         case KeyEvent.VK_PAGE_UP:
         vbar.setValue(vbar.getValue() - vbar.getBlockIncrement());
         break;

         case KeyEvent.VK_PAGE_DOWN:
         vbar.setValue(vbar.getValue() + vbar.getBlockIncrement());
         break;

         case KeyEvent.VK_UP:
         vbar.setValue(vbar.getValue() - vbar.getUnitIncrement());
         break;

         case KeyEvent.VK_DOWN:
         vbar.setValue(vbar.getValue() + vbar.getUnitIncrement());
         break;

         case KeyEvent.VK_LEFT:
         hbar.setValue(hbar.getValue() - hbar.getUnitIncrement());
         break;

         case KeyEvent.VK_RIGHT:
         hbar.setValue(hbar.getValue() + hbar.getUnitIncrement());
         break;
         }
      }
      
   public void keyReleased(KeyEvent e) {} //- implements KeyListener
   public void keyTyped(KeyEvent e)    {} //- implements KeyListener

   synchronized
   public void
   scrollbarTouched(TextScrollbar source) //- implements TextScrollbarListener
      {
      canvas.requestFocus(); // start sending keystokes to text area now
      }

   synchronized
   public void
   scrollbarChanged(TextScrollbar source) //- implements TextScrollbarListener
      {
   // if (source == vbar) canvas.setTopmostLine((int)(source.getValue() + 0.5));
   // if (source == hbar) canvas.setLeftmostColumn((int)(source.getValue() + 0.5));
   
      int val = double2int(source.getValue());
      if (source == vbar) canvas.setTopmostLine(val);
      if (source == hbar) canvas.setLeftmostColumn(val);
   
      if (canvas.needsRedraw && isShowing())
         {
         Util.redraw(canvas);
         canvas.needsRedraw = false;
         }
      }

   // Render ourself.
   //
   synchronized
   public void
   paint(Graphics g) //- extends Component
      {
      Rectangle localBounds = Util.getLocalBounds(this);
     
      // paint background
      //
      g.setColor(getBackground());
      g.fillRect(0, 0, localBounds.width, localBounds.height);

      // paint edging, if scrollbar(s) are in use or edging decoration was requested
      //
      if (vbar.getSize().width != 0 || hbar.getSize().height != 0 || edgingStyle != NO_EDGING)
         switch (edgingStyle)
            {
            case NO_EDGING:
            break;
            
            case INSET_EDGING:
            Util.drawInsetBorder(g, localBounds);
            break;

            case OUTSET_EDGING:
            Util.drawOutsetBorder(g, localBounds);
            break;
            }
      
      // paint interior components
      //
      for (int i = getComponentCount(); --i >= 0; )
         {
         Component c = getComponent(i);

         if (!c.isVisible())
            continue; // component is not to be rendered

         Rectangle r = c.getBounds();

         if (!r.intersects(localBounds))
            continue; // component is completely outside my borders

         Shape s = g.getClip();
         g.translate(r.x, r.y);
         g.clipRect(0, 0, r.width, r.height);
         c.paint(g);
         g.translate(-r.x, -r.y);
         g.setClip(s);
         }
      }
   
   // workaround for bug in Solaris implementation of d2i bytecode
   //
   private static final int double2int(double d) { return Double.isNaN(d) ? 0 : (int)d; }
   }

// A mouse- and keystroke- sensitive canvas painted with lines of text.
// If the canvas is larger than the window in which it is displayed, it can be scrolled with the cursor keys.
//
class TextCanvas extends FastComponent
   {
   private String  buffer;         // text to be rendered
   private int     start[];        // index of start of each line in "buffer", plus an extra for dummy "eof" line

                                   // text dimensions...
           int     columns;        // ...number of characters in longest line
           int     lines;          // ...total number of lines

                                   // font dimensions...
           int     fontW;          // ...character column width (pixels)
           int     fontH;          // ...character line height  (pixels)
           int     fontAscent;     // ...distance from top of character to baseline    (pixels)
           int     fontDescent;    // ...distance from baseline to bottom of character (pixels)
   
                                   // scrolling position...
           int     leftmostColumn; // ...column number to appear at left of canvas
           int     topmostLine;    // ...line number to appear at top of canvas

                                   // marking position...
   private int     markedLine;     // ...line number to appear "marked" (-1 --> none)

   private Image   decorations[];  // decorations, one for each line

           Vector  listeners;      // listeners that wish to be notified of mouse clicks
           
           boolean needsRedraw;    // has internal state changed?
          
   // Create an empty canvas.
   //
   TextCanvas()
      {
      this.listeners = new Vector();
      needsRedraw    = true;
      setText("");
      setFont(new Font("Monospaced", Font.PLAIN, 14));
      addMouseListener(new MouseAdapter() { public void mousePressed(MouseEvent e) { TextCanvas.this.deliverClick(e); }});
      }

  // Specify a font (it must be monospaced, in order that horizontal scrolling work correctly).
  //
  public void
  setFont(Font font) //- overrides Component
     {
  // System.out.println("setFont " + getName());
     super.setFont(font);
     FontMetrics metrics  = getFontMetrics(getFont());

     fontW       = metrics.charWidth('X');
     fontH       = metrics.getHeight();
     fontAscent  = metrics.getMaxAscent();
     fontDescent = metrics.getDescent();
     fontH -= fontDescent; // this gives nicer screen appearance (less whitespace between lines)
     }
   
   // Specify client to be informed of mouse clicks.
   //
   synchronized
   void 
   addTextListener(TextListener listener)
      {
      listeners.addElement(listener);
      }

   // Specify text to appear on canvas.
   // Taken: character string with lines delimited by \n characters.
   //
   synchronized
   void
   setText(String text)
      {
   // System.out.println("setText " + getName());

      buffer  = text;
      columns = 0;
      lines   = 0;

      // measure text dimensions
      for (int startOfLine = 0;;)
         {
         int endOfLine = buffer.indexOf('\n', startOfLine);  // (\n == lineSep)
         if (endOfLine < 0) break;
         int cols = endOfLine - startOfLine;
         if (cols > columns)
            columns = cols;
         startOfLine = endOfLine + 1;                        // (1==lineSep.length)
         lines += 1;
         }

      // remember each line's starting index, adding an extra one for dummy "eof" line
      start = new int[lines + 1];
      int lino = 0;
      for (int startOfLine = 0; lino < lines; )
         {
         start[lino++] = startOfLine;
         startOfLine = buffer.indexOf('\n', startOfLine) + 1; // (\n==lineSep, 1==lineSep.length)
         }
      start[lines] = buffer.length();
      
      // set scrolling position
      topmostLine = leftmostColumn = 0;
      
      // set marking position
      markedLine = -1;
      
      // initialize (empty) list of decorations
      decorations = new Image[lines];
      
      needsRedraw = true;
      }

   // Specify line to appear topmost on canvas.
   //
   synchronized
   void
   setTopmostLine(int lino)
      {
      if (lines == 0)
         return;

      int original = topmostLine;
      topmostLine = lino;
      if (topmostLine < 0)
         topmostLine = 0;
      if (topmostLine > lines - 1)
         topmostLine = lines - 1;
      if (topmostLine == original)
         return;
      needsRedraw = true;
      }
      
   // Specify column to appear leftmost on canvas.
   // Returned: did anything actually change?
   //
   synchronized
   void
   setLeftmostColumn(int colno)
      {
      if (columns == 0)
         return;

      int original = leftmostColumn;
      leftmostColumn = colno;
      if (leftmostColumn < 0)
         leftmostColumn = 0;
      if (leftmostColumn > columns - 1)
         leftmostColumn = columns - 1;
      if (leftmostColumn == original)
         return;
      needsRedraw = true;
      }

   // Calculate value of "topmostLine" required to bring "lino" into view.
   // Returned: -1        --> "lino" is currently in view
   //           otherwise --> value of "topmostLine" required to bring "lino" into view at center of window.
   //
   synchronized
   int
   getViewpointAdjustmentValue(int lino)
      {
      int visibleLines   = getSize().height / fontH;
      int bottommostLine = topmostLine + visibleLines - 1;
      
      if (topmostLine <= lino && lino <= bottommostLine) 
         return -1; // already in view
      
      int newtop = lino - visibleLines / 2;
      if (newtop < 0) 
         newtop = 0;

      return newtop; // value required to bring line to center of view
      }

   // Specify line to be hilighted.
   //
   synchronized
   void
   setMarkedLine(int lino)
      {
      if (markedLine == lino)
         return;
      markedLine = lino;
      needsRedraw = true;
      return;
      }
      
   // Specify line to be decorated.
   //
   synchronized
   void
   setDecoratedLine(int lino, Image decoration)
      {
      if (lino < 0 || lino >= decorations.length || decorations[lino] == decoration)
         return;
      decorations[lino] = decoration;
      needsRedraw = true;
      return;
      }

   // Remove all decorations.
   //
   synchronized
   void
   undecorate()
      {
      decorations = new Image[lines];
      needsRedraw = true;
      }

   // Search for a string.
   // Taken:    line at which to begin search
   //           string sought
   //           direction to search (+1 --> forward, -1 --> backward)
   // Returned: line at which string was found (-1 --> not found)
   //
   synchronized
   int
   search(int lino, String sought, int direction)
      {
      for (;;)
         {
         lino += direction;
         if (lino < 0 || lino >= lines)
            return -1;
         int    beg  = start[lino];
         int    end  = start[lino + 1] - 1; // (1==lineSep.length)
         String line = buffer.substring(beg, end);
         if (line.indexOf(sought) != -1)
            return lino;
         }
      }
      
   synchronized
   public void
   paint(Graphics g) //- overrides Component
      {
   // System.out.println("paintText " + getName());
      
      int x = 0;
      int y = 0 + fontAscent;
      int canvasWidth  = getSize().width;
      int canvasHeight = getSize().height;
     
      g.setColor(getBackground());
      g.fillRect(0, 0, canvasWidth, canvasHeight);
      g.setColor(getForeground());
      g.setFont(getFont());

      for (int lino = topmostLine; lino < lines; ++lino)
         {
         int beg = start[lino] + leftmostColumn;
         int end = start[lino + 1] - 1; // (1==lineSep.length)
         if (beg > end)
            beg = end;

         // draw hilights
         if (lino == markedLine)
            {
            Color c = g.getColor();
            g.setColor(Color.white);
            g.fillRect(x, y - fontAscent + fontDescent + 1, 999999, fontH);
            g.setColor(c);
            }

         // draw text
         g.drawString(buffer.substring(beg, end), x, y);

         // draw decoration at left margin
         if (decorations[lino] != null)
            g.drawImage(decorations[lino], x, y - fontAscent + fontDescent + 1, null);

         // advance
         y += fontH;
         if (y - fontAscent >= canvasHeight)
            break;
         }
      }

   synchronized
   void
   deliverClick(MouseEvent e)
      {
      int x = e.getX();
      int y = e.getY();
      
      // right mouse button
      //
      if ((e.getModifiers() & InputEvent.BUTTON3_MASK) != 0)
         { 
         for (int i = 0, n = listeners.size(); i < n; ++i)
            {
            TextListener listener = (TextListener)listeners.elementAt(i);
            listener.textMetaClicked(this, x, y);
            }
         return;
         }
      
      // left mouse button
      //
      int lino  = translateToLineNumber(y);
      int colno = translateToColumnNumber(x);
      if (lino < lines)
         {
         int beg = start[lino];
         int end = start[lino + 1] - 1; // (1==lineSep.length)
         if (beg > end)
            beg = end;

         String text = buffer.substring(beg, end);
         for (int i = 0, n = listeners.size(); i < n; ++i)
            {
            TextListener listener = (TextListener)listeners.elementAt(i);
            listener.textClicked(text, lino, colno);
            }
         
         requestFocus(); // start sending keystokes our way now
         }
      }

   // Translate pixel coordinate `y' to corresponding line number.
   //
   synchronized
   private int
   translateToLineNumber(int y)
      {
      return topmostLine + (y - fontDescent - 2) / fontH;
      }
  
   // Translate pixel coordinate `x' to corresponding column number.
   //
   synchronized
   private int
   translateToColumnNumber(int x)
      {
      return leftmostColumn + x / fontW;
      }
   }

interface TextScrollbarListener
   {
   void scrollbarTouched(TextScrollbar source);
   void scrollbarChanged(TextScrollbar source);
   }

// A standard scrollbar.
//
// Example: isVertical=false valMin=0 valCur=2 valMax=2 valVis=1
//
//            |<----tray----->|
// button          bounds          button
//     \ +--+ +---+---+---+---+ +--+/
//       |< | | 0 | 1 | 2 |###| | >|
//       +--+ +---+---+---+---\ +--+
//            |<--motion->|    thumb
//                bounds
//
class TextScrollbar extends FastContainer
   implements Constants, MouseListener, MouseMotionListener, ActionListener, AlarmListener
   {
   // orientation
   //
   private boolean isVertical;
   
   // up/down or left/right buttons
   //
   private MultiButton button0;
   private MultiButton button1;

   // thumb location, world units
   //
   private double valMin         = 0;
   private double valCur         = 0;
   private double valMax         = 0;
   private double valVis         = 0;
   private double unitIncrement  = 1;
   private double blockIncrement = 1;
   private double pageIncrement;

   // thumb location, graphics units (pixels)
   //
   private double  thumbX;
   private double  thumbY;
   private double  thumbW;
   private double  thumbH;
   private boolean thumbPressed;
   private boolean thumbDragging;
   private double  thumbDragStartX;
   private double  thumbDragStartY;
   
   // scrollbar "tray" location
   //
   private double trayX;
   private double trayY;
   private double trayH;
   private double trayW;
   
   // listeners on this scrollbar
   //
   private Vector listeners;
   
   // autorepeat event generator
   //
   private static AlarmGenerator alarmGenerator;
   
   //-----------//
   // interface //
   //-----------//
   
   TextScrollbar(boolean isVertical)
      { 
      this.isVertical = isVertical;
      
      button0   = new MultiButton(null, null, MultiButton.IS_AUTOREPEATER);
      button1   = new MultiButton(null, null, MultiButton.IS_AUTOREPEATER);
      listeners = new Vector();
      
      add(button0);
      add(button1);
      addMouseListener(this);
      addMouseMotionListener(this);
      
      button0.addActionListener(this);
      button1.addActionListener(this);
      
      if (alarmGenerator == null)
         alarmGenerator = new AlarmGenerator(true);
      }

    // Set scrollbar's thumb-value to world-value mapping characteristics (see example, above).
    // Assumptions: min < max
    //              vis <= (max - min)
    //
    synchronized 
    public void 
    setValues(double valCur, double valVis, double valMin, double valMax)
       {
       this.valVis = valVis;
       this.valMin = valMin;
       this.valMax = valMax;
       
       if (isVertical)
          {
          thumbW = SCROLLBAR_THICKNESS;
          thumbH = valVis * trayH / (valMax - valMin + valVis);
          if (thumbH < MIN_THUMB_SIZE) thumbH = MIN_THUMB_SIZE;
          }
       else
          {
          thumbW = valVis * trayW / (valMax - valMin + valVis);
          thumbH = SCROLLBAR_THICKNESS;
          if (thumbW < MIN_THUMB_SIZE) thumbW = MIN_THUMB_SIZE;
          }
 
       setValue(valCur);
       }
       
   // Listen for adjustments to this scrollbar.
   //
   synchronized
   public void
   addListener(TextScrollbarListener listener)
      {
      listeners.addElement(listener);
      }

   // Change the scrollbar's "thumb" to correspond to a new value.
   //
   synchronized
   void
   setValue(double val)
      {
      valCur = val;
      if (valCur < valMin) valCur = valMin;
      if (valCur > valMax) valCur = valMax;
      updateValue(valCur);
      }

   synchronized
   double
   getValue()
      { 
      return valCur;
      }

   synchronized
   void
   setUnitIncrement(double unitIncrement)
      {
      this.unitIncrement = unitIncrement;
      }
   
   synchronized
   double
   getUnitIncrement()
      {
      return unitIncrement;
      }
   
   synchronized
   void
   setBlockIncrement(int blockIncrement)
      {
      this.blockIncrement = blockIncrement;
      }
   
   synchronized
   double
   getBlockIncrement()
      {
      return blockIncrement;
      }

   //------------------------//
   // implementation details //
   //------------------------//

   synchronized
   public void
   setBounds(int x, int y, int width, int height) //- extends Component
      {
      Dimension old = getSize();
      super.setBounds(x, y, width, height);

      if (old.width == width && old.height == height)
         return;

      double buttonW = SCROLLBAR_THICKNESS;
      double buttonH = SCROLLBAR_THICKNESS;
      
      if (isVertical)
         {
         button0.setBounds(0, 0, (int)buttonW, (int)buttonH);
         trayX = 0;
         trayY = buttonH;
         trayW = width;
         trayH = height - 2 * buttonH;
         button1.setBounds(0, (int)(buttonH + trayH), (int)buttonW, (int)buttonH);
         thumbH = valVis * trayH / (valMax - valMin + valVis);
         if (thumbH < MIN_THUMB_SIZE) thumbH = MIN_THUMB_SIZE;
         }
      else
         {
         button0.setBounds(0, 0, (int)buttonW, (int)buttonH);
         trayX = buttonW;
         trayY = 0;
         trayW = width - 2 * buttonW;
         trayH = height;
         button1.setBounds((int)(buttonW + trayW), 0, (int)buttonW, (int)buttonH);
         thumbW = valVis * trayW / (valMax - valMin + valVis);
         if (thumbW < MIN_THUMB_SIZE) thumbW = MIN_THUMB_SIZE;
         }
      }

   synchronized
   public void
   actionPerformed(ActionEvent e) //- implements ActionListener
      { // scrollbar button pressed
      if (e.getSource() == button0) updateValue(valCur - unitIncrement);
      if (e.getSource() == button1) updateValue(valCur + unitIncrement);

      if (e.getActionCommand().equals("click")) // just deliver one "touched" event per button press (ignore autorepeats)
         for (int i = 0, n = listeners.size(); i < n; ++i)
            ((TextScrollbarListener)listeners.elementAt(i)).scrollbarTouched(this);
      }
      
   synchronized
   public void
   mousePressed(MouseEvent e) //- implements MouseListener
      {
      if (thumbPressed) return; // bogus press event
      
      for (int i = 0, n = listeners.size(); i < n; ++i)
         ((TextScrollbarListener)listeners.elementAt(i)).scrollbarTouched(this);
      
      double x = e.getX();
      double y = e.getY();
      double w = getBounds().width;
      double h = getBounds().height;
      
      thumbPressed = true;
      boolean alarmNeeded = true;

      if (isVertical)
         {
         if (y < trayY || y > (trayY + trayH))
            { // button
            }
         else if (y < thumbY)
            { // block back
            pageIncrement = -blockIncrement;
            incrementPage();
            }
         else if (y < thumbY + thumbH)
            { // drag
            thumbDragging   = true;
            thumbDragStartX = x - thumbX;
            thumbDragStartY = y - thumbY;
            alarmNeeded = false;
            }
         else
            { // block forward
            pageIncrement = +blockIncrement;
            incrementPage();
            }
         }
      else
         {
         if (x < trayX || x > (trayX + trayW))
            { // button
            }
         else if (x < thumbX)
            { // block left
            pageIncrement = -blockIncrement;
            incrementPage();
            }
         else if (x < thumbX + thumbW)
            { // drag
            thumbDragging   = true;
            thumbDragStartX = x - thumbX;
            thumbDragStartY = y - thumbY;
            alarmNeeded = false;
            }
         else
            { // block right
            pageIncrement = +blockIncrement;
            incrementPage();
            }
         }
    
      if (alarmNeeded)
         alarmGenerator.on(this);
      }
   
   synchronized
   public void
   mouseDragged(MouseEvent e) //- implements MouseMotionListener
      {
      if (!thumbDragging) return;

      double x = e.getX();
      double y = e.getY();

      if (isVertical) updateThumb(y - thumbDragStartY);
      else            updateThumb(x - thumbDragStartX);
      }
   
   synchronized
   public void
   mouseReleased(MouseEvent e) //- implements MouseListener
      {
      if (!thumbPressed) return; // bogus release event

      double x = e.getX();
      double y = e.getY();

      if (thumbDragging)
         {
         if (isVertical) updateThumb(y - thumbDragStartY);
         else            updateThumb(x - thumbDragStartX);
         }

      thumbPressed  = false;
      thumbDragging = false;
     
      alarmGenerator.off();
      }

   synchronized
   public void
   mouseMoved(MouseEvent e) //- implements MouseMotionListener
      {
      }

   public void mouseEntered( MouseEvent e) { } //- implements MouseListener
   public void mouseExited(  MouseEvent e) { } //- implements MouseListener
   public void mouseClicked( MouseEvent e) { } //- implements MouseListener
   
   public void
   alarmFired(AlarmGenerator source) //- implements AlarmListener
      {
      incrementPage();
      }

   synchronized
   void incrementPage()
      {
      updateValue(valCur + pageIncrement);
      }

   synchronized
   private void
   updateValue(double val)
      {
      if (isVertical) updateThumb(trayY + (val - valMin) * (trayH - thumbH) / (valMax - valMin));
      else            updateThumb(trayX + (val - valMin) * (trayW - thumbW) / (valMax - valMin));
      }

   synchronized
   private void
   updateThumb(double newLocation)
      {
      // update state
      //
      if (isVertical)
         {
         thumbY = newLocation;

         // snug bottom
         if (thumbY + thumbH > trayY + trayH)
            thumbY = trayY + trayH - thumbH;

         // snug top
         if (thumbY < trayY)
            thumbY = trayY;

         valCur = valMin + (thumbY - trayY) * (valMax - valMin) / (trayH - thumbH);
         }
      else
         {
         thumbX = newLocation;

         // snug right
         if (thumbX + thumbW > trayX + trayW)
            thumbX = trayX + trayW - thumbW;

         // snug left
         if (thumbX < trayX)
            thumbX = trayX;

         valCur = valMin + (thumbX - trayX) * (valMax - valMin) / (trayW - thumbW);
         }

      // update screen
      //
      if (isShowing())
         {
         int      w     = getSize().width;
         int      h     = getSize().height;
         Image    image = null;
         Graphics ig    = null;
         Graphics g     = null;
         try
            {
            image = createImage(w, h);
            ig    = image.getGraphics();
            g     = getGraphics();

            draw(ig);                       // draw ourself onto offscreen image
            g.drawImage(image, 0, 0, null); // copy offscreen image onto screen
            }
         catch(Exception e)
            {
            if (image != null) image.flush();
            if (ig    != null) ig.dispose();
            if (g     != null) g.dispose();
            }
         }

      // update listeners
      //
      for (int i = 0, n = listeners.size(); i < n; ++i)
         ((TextScrollbarListener)listeners.elementAt(i)).scrollbarChanged(this);
      }

   synchronized
   private void
   draw(Graphics g)
      {
      // buttons
      //
      Point p = button0.getLocation();
      g.translate(p.x, p.y);
      button0.paint(g);
      g.translate(-p.x, -p.y);
      
      p = button1.getLocation();
      g.translate(p.x, p.y);
      button1.paint(g);
      g.translate(-p.x, -p.y);

      // tray background
      //
      g.setColor(SCROLLBAR_BACKGROUND);
      g.fillRect((int)trayX, (int)trayY, (int)trayW, (int)trayH);
      
      int x = (int)thumbX;
      int y = (int)thumbY;
      int w = (int)thumbW;
      int h = (int)thumbH;
      
      // thumb outline
      //
      Util.drawOutsetBorder(g, new Rectangle(x, y, w, h));
      
      // thumb interior
      //
      g.setColor(SystemColor.controlHighlight);
      g.fillRect(x + BEVEL_THICKNESS, y + BEVEL_THICKNESS, w - 2 * BEVEL_THICKNESS, h - 2 * BEVEL_THICKNESS);
      }

   synchronized
   public void
   paint(Graphics g) //- overrides Component
      {
      draw(g);
      }
   }
